# vim: tabstop=4 shiftwidth=4 softtabstop=4

# Copyright (c) 2011 NTT.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import socket
from zope import interface

from nova import flags
from nova import rpc
from nova import utils
from nova.network import service
from nova.network.flat import common
from nova.network.flat.db import api as db_api

FLAGS = flags.FLAGS
flags.DEFINE_string('ns_flat_network_host', socket.gethostname(),
                    'Network host to use for ip allocation in flat modes')
flags.DEFINE_string('ns_flat_network_topic', 'net-flat', 
                    'the topic network nodes listen on')
flags.DEFINE_string('ns_flat_network_bridge', 'br100',
                    'Bridge for simple network instances')
flags.DEFINE_bool('ns_flat_network_fake_call', False,
                  'If True, skip using the queue and make local calls')

def _queue_get_for(topic, physical_node_id):
    return "%s.%s" % (topic, physical_node_id)

def _get_network_topic(context, **kwargs):
    """Retrieves the network host for a project on this host"""
    if FLAGS.stub_network:
        host = FLAGS.ns_flat_network_host
    else:
        host = _get_network_host(context)
    return _queue_get_for(FLAGS.ns_flat_network_topic,
                          host)

def _get_network_host(context):
    """Get the network host for the current context."""
    network_ref = db_api.network_get_by_bridge(context,
                                               FLAGS.ns_flat_network_bridge)
    host = network_ref['host']
    if not host:
        topic = _queue_get_for(FLAGS.ns_flat_network_topic,
                               FLAGS.ns_flat_network_host)
        if FLAGS.ns_flat_network_fake_call:
            return _set_network_host(context, network_ref['id'])
        host = rpc.call(context,
                        FLAGS.ns_flat_network_topic,
                        {"method": "set_network_host",
                         "args": {"network_id": network_ref['id']}})
    return host

def _set_network_host(context, network_id):
    """Safely sets the host of the network."""
    db_api.network_update(context,
                          network_id,
                          {"host":FLAGS.host,
                           "dns": FLAGS.ns_flat_dns})
    return FLAGS.host 

class NetworkAgentService(object):
    """An OpenStack network agent service object."""

    interface.implements(service.INetworkOpenStackApiService)
    
    def bind_vnics_to_ports(self, context, vnic_ids, _is_vpn): 
        """Gets a fixed ips from the pool.
        
        Args:
           context: Nova context needed for DB access.
           vnic_ids: list of VNIC IDs
           is_vpn: Boolean to check if the call is for VPN.          
        """
        rpc.call(context,
                 _get_network_topic(context),
                 {"method": "allocate_fixed_ips",
                  "args": {"vnic_ids": vnic_ids}})
    
    def setup_compute_network(self, context, vnic_ids): 
        """Nothing needs to be done.
        """
        pass
    
    def teardown_compute_network(self, context, vnic_ids):
        """Clean up the compute node.
        
        Args:
           context: Nova context needed for DB access.
           vnic_ids: list of VNIC IDs
        """
        ctxt = context.elevated() 
        for vnic_id in vnic_ids:
            ip = vnic_id.get('fixed_ip')
            if not FLAGS.stub_network and ip:
    
                address = ip['address']
                db_api.ip_address_update(ctxt, address, 
                                              {'allocated': False})
                db_api.ip_address_disassociate(ctxt, address)
    
    def requires_file_injection(self):
        """This service requires file injection.
        """
        return True
    
    def get_network_info(self, context, vnic_id):
        """Gets network data for a given VNIC.
        
        Args:
           context: Nova context needed for DB access.
           vnic_id: VNIC ID to get the network information for.
        
        Returns:
           A dictionary containing the following keys:
           cidr, cidrv6, netmask, netmask_v6 gateway, gateway_v6, 
           dhcp_server, broadcast, dns, ra_server, mac_address,
           ip_address, bridge, and address_v6
        """
        ethernet_card = db_api.ethernet_card_get(context, vnic_id)
        if not ethernet_card:
            raise ValueError("Invalid vnic ID!")
    
        # Get the IP address associated with this VNIC
        ip_address = db_api.ip_address_get_by_ethernet_card(context, 
                                                            ethernet_card.id)
        if not ip_address:
            raise ValueError("VNIC is not mapped to IP!")
    
        network = ip_address.network
        if network['cidr_v6']:
            ipv6_addr = utils.to_global_ipv6(network['cidr_v6'],
                                             ethernet_card['mac_address'])
        else:
            ipv6_addr = None
    
        net, mask = common.get_net_and_mask(network['cidr'])
        return dict(mac_address=ethernet_card['mac_address'],
                    ip_address=ip_address['address'], bridge=network['bridge'],
                    cidr=network['cidr'], netmask=network['netmask'],
                    gateway=network['gateway'], broadcast=network['broadcast'],
                    dns=network['dns'], dhcp_server=None, 
                    cidr_v6=network['cidr_v6'], address_v6=ipv6_addr, 
                    ra_server=None, netmask_v6=None, gateway_v6=None,
                    net=net, mask=mask) 

