# vim: tabstop=4 shiftwidth=4 softtabstop=4

# Copyright (c) 2011 NTT.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Wrapper APIs for DB models should be defined here.
"""
import IPy
import math

from nova import flags
from nova import utils

from nova.network.flat.db import api

FLAGS = flags.FLAGS
flags.DEFINE_string('ns_flat_bridge',
                    'br100',
                    'Bridge for simple network instances')

BOTTOM_RESERVED_IPS = 2
TOP_RESERVED_IPS = 1

def _create_fixed_ips(context, network_id):
    """Create all fixed ips for network."""
    network_ref = api.network_get(context, network_id)
    net = IPy.IP(network_ref['cidr'])
    num_ips = len(net)
    for index in range(num_ips):
        address = str(net[index])
        if index < BOTTOM_RESERVED_IPS or num_ips - index < TOP_RESERVED_IPS:
            reserved = True
        else:
            reserved = False

        api.ip_address_create(context, 
                              {'network_id': network_id,
                                'address': address,
                                'reserved': reserved})

def network_create(context, cidr, num_networks, network_size,
           cidr_v6, *args, **kwargs):
    """Create networks based on parameters."""
    fixed_net = IPy.IP(cidr)
    fixed_net_v6 = IPy.IP(cidr_v6)
    significant_bits_v6 = 64
    networks = []
    for index in range(num_networks):
        start = index * network_size
        significant_bits = 32 - int(math.log(network_size, 2))
        cidr = "%s/%s" % (fixed_net[start], significant_bits)
        project_net = IPy.IP(cidr)
        net = {}
        net['bridge'] = FLAGS.ns_flat_bridge
        net['cidr'] = cidr
        net['netmask'] = str(project_net.netmask())
        net['gateway'] = str(project_net[1])
        net['broadcast'] = str(project_net.broadcast())

        if(FLAGS.use_ipv6):
            cidr_v6 = "%s/%s" % (fixed_net_v6[0], significant_bits_v6)
            net['cidr_v6'] = cidr_v6

        network_ref = api.network_create_safe(context, net)

        if network_ref:
            networks.append(network_ref)
            _create_fixed_ips(context, network_ref['id'])

    return networks

def allocate_ip(context, ethernet_card_id, *args, **kwargs):
    """Gets an IP from the pool."""
    network_ref = api.network_get_by_bridge(context,
                                            FLAGS.ns_flat_bridge)
    ip = api.ip_address_associate_pool(context.elevated(),
                                       network_ref['id'],
                                       ethernet_card_id)
    api.ip_address_update(context, ip.id, {'allocated': True})
    return ip

def bind_ip_to_ethernet_card(context, ethernet_card_id):
    """Assigns IP to ethernet card."""
    
    # Get the ethernet card
    ethernet_card = api.ethernet_card_get(context, ethernet_card_id)
    # TODO: Throw a better exception.
    if not ethernet_card:
        raise ValueError("There is not VNIC with ID %s" % ethernet_card_id)

    ip_address = api.ip_address_get_by_ethernet_card(context, 
                                                     ethernet_card_id)
    # Check to see if this ethernet card does not have IP allocated.
    if not ip_address:
        ip_address = allocate_ip(context, ethernet_card_id)

    return ip_address

def network_update_host(context, id, host):
    """Updates network host"""
    return api.network_update(context, id, dict(host=host))

def ethernet_card_create_with_random_mac(context):
    """Creates a new ethernet card"""
    mac_address = utils.generate_mac()
    return api.ethernet_card_create(context, dict(mac_address=mac_address))

    
