# vim: tabstop=4 shiftwidth=4 softtabstop=4

# Copyright (c) 2011 NTT.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""
Implementation of SQLAlchemy backend.
"""
from sqlalchemy import or_
from sqlalchemy.exc import IntegrityError
import warnings

from nova import exception
from nova.db import api as nova_db
from nova.db.sqlalchemy import api as nova_api
from nova.network.flat.db.sqlalchemy import models
from nova.network.flat.db.sqlalchemy.session import get_session


class DataAccessCache(object):
    """A cache of Data Access Objects specific to sessions and contexts.

    Creates one DAO object per session and set of context parameter
    values, and caches it. The parameters currently used in the
    context are:
        - nova.db.sqlalchemy.api.is_admin_context(context)
        - nova.db.sqlalchemy.api.can_read_deleted(context)
    """

    def __init__(self, dao_class):
        self._dao_class = dao_class
        self._cached_daos = {}

    def get_dao(self, context, session):
        """Get a DataAccess object.

        If no cached DAO has been created for this session and this
        context's parameter, a new DAO is created and
        cached. Otherwise, the cached DAO is returned.

        :param context: The request context.
        :param session: The SQLAlchemy session to use as the Data
            Access Object's data source.
        """
        is_admin = nova_api.is_admin_context(context)
        can_read_deleted = nova_api.can_read_deleted(context)
        key = (session, is_admin, can_read_deleted)
        dao = self._cached_daos.get(key)
        if dao is None:
            dao = self._dao_class(session, is_admin, can_read_deleted)
            self._cached_daos[key] = dao
        return dao


class DefaultSessionFlatNetworkDataAccessFactory(object):
    """A DAO factory that lazily uses the default session.

    This factory uses the default session for the flat network
    service, which it creates lazily when creating the first DAO.
    """

    def __init__(self):
        self._dao_cache = DataAccessCache(models.FlatNetworkDataAccess)
        self._session = None

    def get_session(self):
        if self._session is None:
            self._session = get_session()
        return self._session

    def get_dao(self, context):
        """Get a DataAccess object.

        If no cached DAO has been created for this context's
        parameter, a new DAO is created and cached. Otherwise, the
        cached DAO is returned.

        :param context: The request context.
        """
        return self._dao_cache.get_dao(context, self.get_session())


dao_factory = DefaultSessionFlatNetworkDataAccessFactory()
