# vim: tabstop=4 shiftwidth=4 softtabstop=4

# Copyright (c) 2011 NTT.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""
SQLAlchemy models for flat network service data.
"""
from sqlalchemy import or_

from sqlalchemy.orm import relationship, backref
from sqlalchemy.orm import joinedload
from sqlalchemy import Column, Boolean, Integer, String
from sqlalchemy import ForeignKey
from sqlalchemy.ext.declarative import declarative_base

from nova import exception
from nova.db.sqlalchemy.models import NovaBase


BASE = declarative_base()


# ----------------------
# Data Transfer Objects.
# ----------------------

class EthernetCard(BASE, NovaBase):
    """Represents an ethernet card."""
    __tablename__ = 'ethernet_cards'
    id = Column(Integer, primary_key=True)
    mac_address = Column(String(255), nullable=False, unique=True)


class Network(BASE, NovaBase):
    """Represents a simple IP network."""
    __tablename__ = 'networks'
    id = Column(Integer, primary_key=True)
    cidr = Column(String(255))
    netmask = Column(String(255))
    bridge = Column(String(255))
    gateway = Column(String(255))
    broadcast = Column(String(255))
    dns = Column(String(255))
    cidr_v6 = Column(String(255), unique=True)
    host = Column(String(255))  # , ForeignKey('hosts.id'))

    
class IpAddress(BASE, NovaBase):
    """Represents a ip address for an ethernet card."""
    __tablename__ = 'ip_addresses'
    id = Column(Integer, primary_key=True)
    address = Column(String(255))
    network_id = Column(Integer, ForeignKey('networks.id'), nullable=True)
    network = relationship(Network, backref=backref('ip_addresses'))
    allocated = Column(Boolean, default=False)
    reserved = Column(Boolean, default=False)
    ethernet_card_id = Column(Integer, ForeignKey('ethernet_cards.id'))
    ethernet_card = relationship(EthernetCard, backref=backref('ip_addresses'))
    addressV6 = Column(String(255))
    netmaskV6 = Column(String(3))
    gatewayV6 = Column(String(255))


# ----------------------
# Data Access Objects.
# ----------------------

class DataAccess(object):
    """The base class to implement Data Access Objects.
    """

    def __init__(self, session, is_admin, can_read_deleted):
        """Initialize this Data Access Object.

        :param session: The SQLAlchemy session to use as this Data
            Access Object's data source.
        :param is_admin: Indicates if the request context is an
            administrator.
        :param can_read_deleted: Indicates if the request context has
            access to deleted objects.
        """
        self._session = session
        self._is_admin = is_admin
        self._can_read_deleted = can_read_deleted


class EthernetCardDataAccess(DataAccess):
    """A Data Access Object to access ethernet cards.
    """

    ethernet_card_dto_class = EthernetCard

    def ethernet_card_get(self, id):
        result = self._session.query(self.ethernet_card_dto_class).\
            filter_by(deleted=False).\
            filter_by(id=id).\
            first()

        if not result:
            raise exception.NotFound(_("No ethernet_card with id %s") % id)

        return result

    def ethernet_card_get_all(self):
        return self._session.query(self.ethernet_card_dto_class).\
            filter_by(deleted=self._can_read_deleted).\
            all()

    def ethernet_card_create(self, values):
        ethernet_card_ref = self.ethernet_card_dto_class()
        ethernet_card_ref.update(values)
        ethernet_card_ref.save(session=self._session)
        return ethernet_card_ref

    def ethernet_card_update(self, ethernet_card_id, values):
        with self._session.begin():
            ethernet_card_ref = self.ethernet_card_get(ethernet_card_id)
            ethernet_card_ref.update(values)
            ethernet_card_ref.save(session=self._session)

    def ethernet_card_delete(self, id):
        with self._session.begin():
            ethernet_card_ref = self.ethernet_card_get(id)
            ethernet_card_ref.delete(session=self._session)


class NetworkDataAccess(DataAccess):
    """A Data Access Object to access networks.
    """

    network_dto_class = Network

    def network_get(self, id):
        result = self._session.query(self.network_dto_class).\
            filter_by(deleted=False).\
            filter_by(id=id).\
            first()

        if not result:
            raise exception.NotFound(_("No network with id %s") % id)

        return result

    def network_get_by_bridge(self, bridge):
        if not self._is_admin:
            raise exception.NotAuthorized()
        result = self._session.query(self.network_dto_class).\
            filter_by(bridge=bridge).\
            filter_by(deleted=False).\
            first()

        if not result:
            raise exception.NotFound(_('No network for bridge %s') % bridge)

        return result

    def network_get_by_ethernet_card(self, ethernet_card_id):
        if not self._is_admin:
            raise exception.NotAuthorized()
        rv = self._session.query(self.network_dto_class).\
            filter_by(deleted=False).\
            join(self.network_dto_class.fixed_ips).\
            filter_by(ethernet_card_id=ethernet_card_id).\
            filter_by(deleted=False).\
            first()
        if not rv:
            raise exception.NotFound(_('No network for ethernet card %s') % 
                                     ethernet_card_id)
        return rv

    def host_get_networks(self, host):
        with self._session.begin():
            return self._session.query(self.network_dto_class).\
                filter_by(deleted=False).\
                filter_by(host=host).\
                all()

    def network_get_all(self):
        return self._session.query(self.network_dto_class).\
            filter_by(deleted=self._can_read_deleted).\
            all()

    def network_create(self, values):
        network_ref = self.network_dto_class()
        network_ref.update(values)
        network_ref.save(session=self._session)
        return network_ref

    def network_create_safe(self, values):
        if not self._is_admin:
            raise exception.NotAuthorized()
        try:
            return self.network_create(values)
        except IntegrityError:
            return None

    def network_update(self, network_id, values):
        with self._session.begin():
            network_ref = self.network_get(network_id)
            network_ref.update(values)
            network_ref.save(session=self._session)

    def network_delete(self, id):
        with self._session.begin():
            network_ref = self.network_get(id)
            network_ref.delete(session=self._session)


class IpAddressDataAccess(DataAccess):
    """A Data Access Object to access IP addresses.
    """

    ip_address_dto_class = IpAddress

    def ip_address_get(self, id):
        result = self._session.query(self.ip_address_dto_class).\
            filter_by(deleted=False).\
            filter_by(id=id).\
            first()

        if not result:
            raise exception.NotFound(_("No ip address with id %s") % id)

        return result

    def ip_address_get_by_ethernet_card(self, ethernet_card_id):
        result = self._session.query(self.ip_address_dto_class).\
            filter_by(deleted=False).\
            filter_by(ethernet_card_id=ethernet_card_id).\
            first()
        return result

    def ip_address_get_all(self):
        return self._session.query(self.ip_address_dto_class).\
            filter_by(deleted=self._can_read_deleted).\
            all()

    def ip_address_create(self, values):
        ip_address_ref = self.ip_address_dto_class()
        ip_address_ref.update(values)
        ip_address_ref.save(session=self._session)
        return ip_address_ref

    def ip_address_update(self, ip_address_id, values):
        with self._session.begin():
            ip_address_ref = self.ip_address_get(ip_address_id)
            ip_address_ref.update(values)
            ip_address_ref.save(session=self._session)   

    def ip_address_delete(self, id):
        with self._session.begin():
            ip_address_ref = self.ip_address_get(id)
            ip_address_ref.delete(session=self._session)

    def ip_address_get_by_address(self, address):
        result = self._session.query(self.ip_address_dto_class).\
                         filter_by(address=address).\
                         filter_by(deleted=self._can_read_deleted).\
                         options(joinedload('network')).\
                         options(joinedload('ethernet_card')).\
                         first()
        if not result:
            raise exception.NotFound(_('No fixed ip for address %s') % address)
    
        return result

    def ip_address_disassociate(self, address):
        with self._session.begin():
            ip_address_ref = self.ip_address_get_by_address(context, address,
                                     session=self._session)
            ip_address_ref.ethernet_card = None
            ip_address.save(session=self._session)


class FlatNetworkDataAccess(EthernetCardDataAccess, NetworkDataAccess,
                            IpAddressDataAccess):
    """A Data Access Object to access all data for the flat network service.

    Also access IP address associations.
    """

    def ip_address_associate_pool(self, network_id, ethernet_card_id):
        if not self._is_admin:
            raise exception.NotAuthorized()
        with self._session.begin():
            network_or_none = or_(
                self.ip_address_dto_class.network_id == network_id,
                self.ip_address_dto_class.network_id == None)
            ip_ref = self._session.query(self.ip_address_dto_class).\
                filter(network_or_none).\
                filter_by(reserved=False).\
                filter_by(deleted=False).\
                filter_by(ethernet_card=None).\
                with_lockmode('update').\
                first()
            if not ip_ref:
                raise nova_db.NoMoreAddresses()
            if not ip_ref.network:
                ip_ref.network = self.network_get(network_id)
            ip_ref.ethernet_card = self.ethernet_card_get(ethernet_card_id)
            self._session.add(ip_ref)
        return ip_ref
