# vim: tabstop=4 shiftwidth=4 softtabstop=4

# Copyright (c) 2011 NTT.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""
SQLAlchemy models for flat DHCP network service data.
"""

from sqlalchemy.orm import joinedload_all
from sqlalchemy.orm import relationship, backref
from sqlalchemy import Column, Boolean, Integer, String
from sqlalchemy import ForeignKey
from sqlalchemy.ext.declarative import declarative_base

from nova.db.sqlalchemy.models import NovaBase
from nova.network.flat.db.sqlalchemy import models as flat_models


BASE = declarative_base()


# ----------------------
# Data Transfer Objects.
# ----------------------

class EthernetCard(BASE, NovaBase):
    """Represents an ethernet card."""
    __tablename__ = 'ethernet_cards'
    id = Column(Integer, primary_key=True)
    mac_address = Column(String(255), nullable=False, unique=True)

class Network(BASE, NovaBase):
    """Represents a simple IP network with DHCP service."""
    __tablename__ = 'networks'
    id = Column(Integer, primary_key=True)
    cidr = Column(String(255))
    netmask = Column(String(255))
    bridge = Column(String(255))
    gateway = Column(String(255))
    broadcast = Column(String(255))
    dns = Column(String(255))
    cidr_v6 = Column(String(255), unique=True)
    host = Column(String(255))  # , ForeignKey('hosts.id'))

    dhcp_start = Column(String(255))
    ra_server = Column(String(255))
    label = Column(String(255))

    
class FixedIp(BASE, NovaBase):
    """Represents a fixed ip for an instance."""
    __tablename__ = 'fixed_ips'
    id = Column(Integer, primary_key=True)
    address = Column(String(255))
    network_id = Column(Integer, ForeignKey('networks.id'), nullable=True)
    network = relationship(Network, backref=backref('fixed_ips'))
    ethernet_card_id = Column(Integer, ForeignKey('ethernet_cards.id'))
    ethernet_card = relationship(EthernetCard,
                                 backref=backref('fixed_ips'))
    allocated = Column(Boolean, default=False)
    leased = Column(Boolean, default=False)
    reserved = Column(Boolean, default=False)
    addressV6 = Column(String(255))
    netmaskV6 = Column(String(3))
    gatewayV6 = Column(String(255))


class FloatingIp(BASE, NovaBase):
    """Represents a floating ip that dynamically forwards to a fixed ip."""
    __tablename__ = 'floating_ips'
    id = Column(Integer, primary_key=True)
    address = Column(String(255))
    fixed_ip_id = Column(Integer, ForeignKey('fixed_ips.id'), nullable=True)
    fixed_ip = relationship(FixedIp,
                            backref=backref('floating_ips'),
                            foreign_keys=fixed_ip_id,
                            primaryjoin='and_('
                                'FloatingIp.fixed_ip_id == FixedIp.id,'
                                'FloatingIp.deleted == False)')
    project_id = Column(String(255))
    host = Column(String(255))  # , ForeignKey('hosts.id'))    


# ----------------------
# Data Access Objects.
# ----------------------

class EthernetCardDataAccess(flat_models.EthernetCardDataAccess):
    """A Data Access Object to access networks.
    """

    ethernet_card_dto_class = EthernetCard


class NetworkDataAccess(flat_models.NetworkDataAccess):
    """A Data Access Object to access networks.
    """

    network_dto_class = Network


class FixedIpDataAccess(flat_models.DataAccess):
    """A Data Access Object to access fixed IPs.
    """

    fixed_ip_dto_class = FixedIp

    def fixed_ip_create(self, values):
        fixed_ip_ref = self.fixed_ip_dto_class()
        fixed_ip_ref.update(values)
        fixed_ip_ref.save(session=self._session)
        return fixed_ip_ref['address']

    def fixed_ip_update(self, fixed_ip_id, values):
        with self._session.begin():
            fixed_ip_ref = self.fixed_ip_get(fixed_ip_id)
            fixed_ip_ref.update(values)
            fixed_ip_ref.save(session=self._session)

    def fixed_ip_get(self, fixed_ip_id):
        result = self._session.query(self.fixed_ip_dto_class).\
            filter_by(deleted=False).\
            filter_by(id=fixed_ip_id).\
            first()

        if not result:
            raise exception.NotFound(_("No Fixed IP with id %s") % id)

        return result

    def fixed_ip_get_all(self):
        result = self._session.query(self.fixed_ip_dto_class).all()
        if not result:
            raise exception.NotFound(_('No fixed ips defined'))

        return result

    def network_get_associated_fixed_ips(self, network_id):
        if not self._is_admin:
            raise exception.NotAuthorized()
        return self._session.query(self.fixed_ip_dto_class).\
            options(joinedload_all('ethernet_card')).\
            filter_by(network_id=network_id).\
            filter(self.fixed_ip_dto_class.ethernet_card_id != None).\
            filter_by(deleted=False).\
            all()

    network_get_associated_ips = network_get_associated_fixed_ips

    def fixed_ip_get_by_address(self, address):
        result = self._session.query(self.fixed_ip_dto_class).\
            filter_by(address=address).\
            filter_by(deleted=self._can_read_deleted).\
            options(joinedload('network')).\
            options(joinedload('ethernet_card')).\
            first()
        if not result:
            raise exception.NotFound(_('No fixed ip for address %s') % address)

        return result

    def fixed_ip_disassociate(self, fixed_ip_id):
        with self._session.begin():
            fixed_ip_ref = self.fixed_ip_get(fixed_ip_id)
            fixed_ip_ref.ethernet_card = None
            fixed_ip_ref.save(session=self._session)

    def fixed_ip_get_by_ethernet_card(self, ethernet_card_id):
        result = self._session.query(self.fixed_ip_dto_class).\
            filter_by(ethernet_card_id=ethernet_card_id).\
            filter_by(deleted=self._can_read_deleted).\
            first()
        return result


class FloatingIpDataAccess(flat_models.DataAccess):
    """A Data Access Object to access floating IPs.
    """

    floating_ip_dto_class = FloatingIp

    def floating_ip_allocate_address(self, host, project_id):
        # TODO: Check authorizations much earlier before calling
        # this. Doing that in DB layer is way too late.
        # nova_api.authorize_project_context(context, project_id)
        with self._session.begin():
            floating_ip_ref = self._session.query(self.floating_ip_dto_class).\
                filter_by(host=host).\
                filter_by(fixed_ip_id=None).\
                filter_by(project_id=None).\
                filter_by(deleted=False).\
                with_lockmode('update').\
                first()
            # NOTE(vish): if with_lockmode isn't supported, as in sqlite,
            #             then this has concurrency issues
            if not floating_ip_ref:
                raise nova_db.NoMoreAddresses()
            floating_ip_ref['project_id'] = project_id
            self._session.add(floating_ip_ref)
        return floating_ip_ref['address']

    def floating_ip_count_by_project(self, project_id):
        # TODO: Check authorizations much earlier before calling
        # this. Doing that in DB layer is way too late.
        # nova_api.authorize_project_context(context, project_id)
        return self._session.query(self.floating_ip_dto_class).\
            filter_by(project_id=project_id).\
            filter_by(deleted=False).\
            count()

    def floating_ip_deallocate(self, address):
        with self._session.begin():
            # TODO(devcamcar): How to ensure floating id belongs to user?
            floating_ip_ref = self.floating_ip_get_by_address(address)
            floating_ip_ref['project_id'] = None
            floating_ip_ref.save(session=self._session)

    def floating_ip_destroy(self, address):
        with self._session.begin():
            # TODO(devcamcar): Ensure address belongs to user.
            floating_ip_ref = self.floating_ip_get_by_address(address)
            floating_ip_ref.delete(session=self._session)

    def floating_ip_disassociate_by_address(self, address):
        with self._session.begin():
            # TODO(devcamcar): Ensure address belongs to user.
            #                  Does get_floating_ip_by_address handle this?
            floating_ip_ref = self.floating_ip_get_by_address(address)
            fixed_ip_ref = floating_ip_ref.fixed_ip
            if fixed_ip_ref:
                fixed_ip_address = fixed_ip_ref['address']
            else:
                fixed_ip_address = None
            floating_ip_ref.fixed_ip = None
            floating_ip_ref.save(session=self._session)
        return fixed_ip_address

    def floating_ip_get_all(self):
        if not self._is_admin:
            raise exception.NotAuthorized()
        return self._session.query(self.floating_ip_dto_class).\
            options(joinedload_all('fixed_ip.ethernet_card')).\
            filter_by(deleted=False).\
            all()

    def floating_ip_get_all_by_project(self, project_id):
        # TODO: Check authorizations much earlier before calling
        # this. Doing that in DB layer is way too late.
        # nova_api.authorize_project_context(context, project_id)
        return self._session.query(self.floating_ip_dto_class).\
            options(joinedload_all('fixed_ip.ethernet_card')).\
            filter_by(project_id=project_id).\
            filter_by(deleted=False).\
            all()

    def floating_ip_get_by_address(self, address):
        # TODO(devcamcar): Ensure the address belongs to user.
        result = session.query(self.floating_ip_dto_class).\
            options(joinedload_all('fixed_ip.network')).\
            filter_by(address=address).\
            filter_by(deleted=self._can_read_deleted).\
            first()
        if not result:
            raise exception.NotFound('No floating ip for address %s' % address)

        return result

    def floating_ip_update(self, address, values):
        with self._session.begin():
            floating_ip_ref = self.floating_ip_get_by_address(address)
            for (key, value) in values.iteritems():
                floating_ip_ref[key] = value
            floating_ip_ref.save(session=self._session)

    def floating_ip_create(self, values):
        floating_ip_ref = self.floating_ip_dto_class()
        floating_ip_ref.update(values)
        floating_ip_ref.save(session=self._session)
        return floating_ip_ref

    def floating_ip_disassociate(self, address):
        with self._session.begin():
            # TODO(devcamcar): Ensure address belongs to user.
            #                  Does get_floating_ip_by_address handle this?
            floating_ip_ref = self.floating_ip_get_by_address(address)
            fixed_ip_ref = floating_ip_ref.fixed_ip
            if fixed_ip_ref:
                fixed_ip_address = fixed_ip_ref['address']
            else:
                fixed_ip_address = None
            floating_ip_ref.fixed_ip = None
            floating_ip_ref.save(session=self._session)
        return fixed_ip_address

    def floating_ip_get_all_by_host(self, host):
        if not self._is_admin:
            raise exception.NotAuthorized()
        return self._session.query(self.floating_ip_dto_class).\
                   options(joinedload_all('fixed_ip.ethernet_card')).\
                   filter_by(host=host).\
                   filter_by(deleted=False).\
                   all()


class FlatDhcpNetworkDataAccess(
    EthernetCardDataAccess, NetworkDataAccess, FixedIpDataAccess,
    FloatingIpDataAccess):
    """A Data Access Object to access all data for the flat DHCP network
    service.

    Also access fixed IP address associations.
    """

    def floating_ip_fixed_ip_associate(self, floating_address, fixed_address):
        with self._session.begin():
            # TODO(devcamcar): How to ensure floating_id belongs to user?
            floating_ip_ref = self.floating_ip_get_by_address(floating_address)
            fixed_ip_ref = self.fixed_ip_get_by_address(fixed_address)
            floating_ip_ref.fixed_ip = fixed_ip_ref
            floating_ip_ref.save(session=self._session)

    def fixed_ip_associate_pool(self, network_id, ethernet_card_id):
        if not self._is_admin:
            raise exception.NotAuthorized()
        with self._session.begin():
            network_or_none = or_(
                self.fixed_ip_dto_class.network_id == network_id,
                self.fixed_ip_dto_class.network_id == None)
            ip_ref = self._session.query(self.fixed_ip_dto_class).\
                filter(network_or_none).\
                filter_by(reserved=False).\
                filter_by(deleted=False).\
                filter_by(ethernet_card=None).\
                with_lockmode('update').\
                first()
            if not ip_ref:
                raise nova_db.NoMoreAddresses()
            if not ip_ref.network:
                ip_ref.network = self.network_get(network_id)
                ip_ref.ethernet_card = self.ethernet_card_get(ethernet_card_id)
                self._session.add(ip_ref)
            return ip_ref
