# vim: tabstop=4 shiftwidth=4 softtabstop=4

# Copyright (c) 2011 NTT.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import collections
import os

from nova import context
from nova import flags
from nova import log as logging
from nova import manager
from nova import utils
from nova.db import api as nova_db_api
from nova.network.flat_dhcp.db import manager as db_manager

def _bin_file(script):
    """Return the absolute path to scipt in the bin directory"""
    return os.path.abspath(os.path.join(__file__, "../../../../bin", script))

FLAGS = flags.FLAGS
flags.DEFINE_string('ns_flat_dhcp_network_driver', 'nova.network.linux_net',
                    'Driver to use for network creation')
flags.DEFINE_string('ns_flat_dhcp_dhcp_driver', 'nova.network.linux_dhcp',
                    'Driver to use for DHCP')
flags.DEFINE_string('ns_flat_dhcp_ra_driver', 'nova.network.linux_ra',
                    'Driver to use for RA')
flags.DEFINE_string('ns_flat_dhcp_network_dhcp_start', '10.0.0.2',
                    'Dhcp start for FlatDhcp')
flags.DEFINE_string('ns_flat_dhcp_interface', None,
                    'FlatDhcp will bridge into this interface if set')
flags.DEFINE_string('ns_flat_dhcp_dhcpbridge', 
                    _bin_file('nova-dhcpbridge-flat-dhcp'),
                    'location of nova-dhcpbridge')
flags.DEFINE_string('ns_flat_dhcp_dhcpbridge_flagfile',
                    '/etc/nova/nova-dhcpbridge.conf',
                    'location of flagfile for dhcpbridge')
flags.DEFINE_bool('ns_flat_dhcp_update_dhcp_on_disassociate', False,
                  'Whether to update dhcp when fixed_ip is disassociated')

LOG = logging.getLogger("nova.net_dhcp")

DhcpHost = collections.namedtuple('DhcpHost', 
                                'mac_address, hostname, ip_address, timestamp')

class NetworkService(manager.Manager):

    def __init__(self, network_driver=None, dhcp_driver=None, ra_driver=None,
                 *args, **kwargs):
        """Initializes flat network service."""
        if not network_driver:
            network_driver = FLAGS.ns_flat_dhcp_network_driver
        if not dhcp_driver:
            dhcp_driver = FLAGS.ns_flat_dhcp_dhcp_driver
        if not ra_driver:
            ra_driver = FLAGS.ns_flat_dhcp_ra_driver

        kwargs['db_driver'] = 'nova.network.flat_dhcp.db.api' 
        self.driver = utils.import_object(network_driver)
        self.dhcp_driver = utils.import_object(dhcp_driver)
        self.ra_driver = utils.import_object(ra_driver)
        super(NetworkService, self).__init__(*args, **kwargs)

    def init_host(self):
        """Do any initialization that needs to be run if this is a
        standalone service.
        """
        self.driver.init_host()
        # Set up networking for the projects for which we're already
        # the designated network host.
        ctxt = context.get_admin_context()
        for network in self.db.host_get_networks(ctxt, self.host):
            self.set_network_host(ctxt, network['id'])
        floating_ips = self.db.floating_ip_get_all_by_host(ctxt,
                                                           self.host)
        for floating_ip in floating_ips:
            if floating_ip.get('fixed_ip', None):
                fixed_address = floating_ip['fixed_ip']['address']
                # NOTE(vish): The False here is because we ignore the case
                #             that the ip is already bound.
                self.driver.bind_floating_ip(floating_ip['address'], False)
                self.driver.ensure_floating_forward(floating_ip['address'],
                                                    fixed_address)
        self.driver.metadata_forward()

    def set_network_host(self, context, network_id):
        """Called when this host becomes the host for a project."""
        net = {}
        net['host'] = self.host
        net['dhcp_start'] = FLAGS.ns_flat_dhcp_network_dhcp_start
        self.db.network_update(context, network_id, net)

        network_ref = self.db.network_get(context, network_id)
        bridge = network_ref['bridge']
        self.driver.ensure_bridge(bridge,
                                  FLAGS.ns_flat_dhcp_interface,
                                  network_ref)
        if not FLAGS.fake_network:
            hosts = self._get_dhcp_hosts(context, network_id) 
            self.dhcp_driver.update_dhcp(bridge, 
                                         network_ref['gateway'], 
                                         network_ref['dhcp_start'], 
                                         hosts,
                                         dhcp_script=FLAGS.\
                                                     ns_flat_dhcp_dhcpbridge)
            if(FLAGS.use_ipv6):
                self.ra_driver.update_ra(bridge,
                                         network_ref['cidr_v6'])
                self.db.network_update(context, network_id,
                                       {"ra_server":
                                        utils.get_my_linklocal(bridge)})

    def _get_dhcp_hosts(self, context, network_id):
        """Get a list containing a network's hosts"""
        hosts = []
        for fixed_ip_ref in self.db.\
                                 network_get_associated_fixed_ips(context,
                                                                  network_id):
            ethernet_card = fixed_ip_ref['ethernet_card']
            instance = nova_db_api.\
                       instance_get_by_virtual_nic(context,
                                                   ethernet_card['id'])
            if instance:
                hosts.append(DhcpHost(mac_address=ethernet_card['mac_address'],
                                      hostname=instance['hostname'],
                                      ip_address=fixed_ip_ref['address'],
                                      timestamp=None))
        return hosts


    def allocate_fixed_ips(self, context, vnic_ids):
        """Gets a fixed ip from the pool."""
        for vnic_id in vnic_ids:
            ip_address = db_manager.bind_ip_to_ethernet_card(context, vnic_id)

            if not FLAGS.fake_network:
                network_id = ip_address['network_id']
                network_ref = self.db.network_get(context, network_id)
                hosts = self._get_dhcp_hosts(context, network_id) 
                self.dhcp_driver.update_dhcp(network_ref['bridge'], 
                                             network_ref['gateway'],
                                             network_ref['dhcp_start'],
                                             hosts, 
                                             dhcp_script=FLAGS.\
                                                       ns_flat_dhcp_dhcpbridge)

    def lease_fixed_ip(self, context, mac, address):
        """Called by dhcp-bridge when ip is leased."""
        LOG.debug(_("Leasing IP %s"), address, context=context)
        fixed_ip_ref = self.db.fixed_ip_get_by_address(context, address)
        ethernet_card = fixed_ip_ref['ethernet_card']
        instance = nova_db_api.instance_get_by_virtual_nic(context, 
                                                           ethernet_card['id'])
        if not instance:
            raise exception.Error(_("IP %s leased that isn't associated") %
                                  address)
    
        if nova_db_api.is_user_context(context):
            # Get the project for this IP
            nova_db_api.authorize_project_context(context, instance.project_id)
    
        if ethernet_card['mac_address'] != mac:
            raise exception.Error(_("IP %(address)s leased to bad"
                    " mac %s vs %s") % (ethernet_card['mac_address'], mac)) 
        now = datetime.datetime.utcnow()
        self.db.fixed_ip_update(context,
                               fixed_ip_ref['id'],
                                {'leased': True,
                                 'updated_at': now})
        if not fixed_ip_ref['allocated']:
            LOG.warn(_("IP %s leased that was already deallocated"), address,
                     context=context)
    
    def get_dhcp_host_leases(self, context, interface):
        """Gets a list of leased hosts."""
        network_ref = self.db.network_get_by_bridge(context, interface)
        hosts = []
        for fixed_ip_ref in self.db.\
                            network_get_associated_fixed_ips(context,
                                                            network_ref['id']):
            ethernet_card = fixed_ip_ref['ethernet_card']
            instance = nova_db_api.\
                       instance_get_by_virtual_nic(context,
                                                   ethernet_card['id'])
    
            if instance['updated_at']:
                timestamp = instance['updated_at']
            else:
                timestamp = instance['created_at']
            
            hosts.append(DhcpHost(mac_address=ethernet_card['mac_address'],
                                  hostname=instance['hostname'],
                                  ip_address=fixed_ip_ref['address'],
                                  timestamp=timestamp))
        return self.dhcp_driver.get_dhcp_leases(hosts)
    
    def release_fixed_ip(self, context, mac, address):
        """Called by dhcp-bridge when ip is released."""
        LOG.debug(_("Releasing IP %s"), address, context=context)
        fixed_ip_ref = self.db.fixed_ip_get_by_address(context, address)
        ethernet_card = fixed_ip_ref['ethernet_card']
        instance = nova_db_api.instance_get_by_virtual_nic(context, 
                                                           ethernet_card['id'])
        if not instance:
            raise exception.Error(_("IP %s released that isn't associated") %
                                  address)

        if nova_db_api.is_user_context(context):
            # Get the project for this IP
            nova_db_api.authorize_project_context(context, instance.project_id)

        if ethernet_card['mac_address'] != mac:
            raise exception.Error(_("IP %(address)s leased to bad"
                    " mac %s vs %s") % (ethernet_card['mac_address'], mac)) 

        if not fixed_ip_ref['leased']:
            LOG.warn(_("IP %s released that was not leased"), address,
                     context=context)
        self.db.fixed_ip_update(context,
                                fixed_ip_ref['address'],
                                {'leased': False})
        if not fixed_ip_ref['allocated']:
            self.db.fixed_ip_disassociate(context, fixed_ip_ref['id'])
            # NOTE(vish): dhcp server isn't updated until next setup, this
            #             means there will stale entries in the conf file
            #             the code below will update the file if necessary
            if FLAGS.ns_flat_dhcp_update_dhcp_on_disassociate:
                network_ref = fixed_ip['network']
                hosts = self._get_dhcp_hosts(context, network_ref['id']) 
                self.dhcp_driver.update_dhcp(network_ref['bridge'], 
                                             network_ref['gateway'], 
                                             network_ref['dhcp_start'], 
                                             hosts,
                                             dhcp_script=FLAGS.\
                                                     ns_flat_dhcp_dhcpbridge)

    def disassociate_floating_ip(self, context, floating_address):
        """Disassociates a floating ip."""
        fixed_address = self.db.floating_ip_disassociate(context,
                                                         floating_address)
        self.driver.unbind_floating_ip(floating_address)
        self.driver.remove_floating_forward(floating_address, fixed_address)

