
# vim: tabstop=4 shiftwidth=4 softtabstop=4

# Copyright (c) 2011 NTT.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import os

from nova import flags
from nova import log as logging
from nova import utils

LOG = logging.getLogger("nova.linux_dhcp")

def _bin_file(script):
    """Return the absolute path to scipt in the bin directory"""
    return os.path.abspath(os.path.join(__file__, "../../../bin", script))

FLAGS = flags.FLAGS
flags.DEFINE_string('linux_dhcp_path', '$state_path/networks',
                    'Location to keep network config files')
flags.DEFINE_string('linux_dhcp_domain', 'novalocal',
                    'domain to use for building the hostnames')
flags.DEFINE_string('linux_dns_server', None,
                    'if set, uses specific dns server for dnsmasq')
flags.DEFINE_string('linux_dhcpbridge_flagfile',
                    '/etc/nova/nova-dhcpbridge.conf',
                    'location of flagfile for dhcpbridge')
flags.DEFINE_string('linux_dhcpbridge', _bin_file('nova-dhcpbridge'),
                    'location of nova-dhcpbridge')

def _dhcp_file(bridge, kind):
    """Return path to a pid, leases or conf file for a bridge"""
    if not os.path.exists(FLAGS.linux_dhcp_path):
        os.makedirs(FLAGS.linux_dhcp_path)
    return os.path.abspath("%s/nova-%s.%s" % (FLAGS.linux_dhcp_path,
                                              bridge,
                                              kind))

def _dnsmasq_cmd(bridge, gateway, dhcp_start, dhcp_script):
    """Builds dnsmasq command"""
    cmd = ['sudo', '-E', 'dnsmasq',
           '--strict-order',
           '--bind-interfaces',
           '--conf-file=',
           '--domain=%s' % FLAGS.linux_dhcp_domain,
           '--pid-file=%s' % _dhcp_file(bridge, 'pid'),
           '--listen-address=%s' % gateway,
           '--except-interface=lo',
           '--dhcp-range=%s,static,120s' % dhcp_start,
           '--dhcp-hostsfile=%s' % _dhcp_file(bridge, 'conf'),
           '--dhcp-script=%s' % dhcp_script, 
           '--leasefile-ro']
    if FLAGS.linux_dns_server:
        cmd += ['-h', '-R', '--server=%s' % FLAGS.linux_dns_server]
    return cmd

def _dnsmasq_pid_for(bridge):
    """Returns the pid for prior dnsmasq instance for a bridge

    Returns None if no pid file exists

    If machine has rebooted pid might be incorrect (caller should check)
    """

    pid_file = _dhcp_file(bridge, 'pid')

    if os.path.exists(pid_file):
        with open(pid_file, 'r') as f:
            return int(f.read())

def _execute(*cmd, **kwargs):
    """Wrapper around utils._execute for fake_network"""
    if FLAGS.fake_network:
        LOG.debug("FAKE NET: %s", " ".join(map(str, cmd)))
        return "fake", 0
    else:
        return utils.execute(*cmd, **kwargs)

def _host_dhcp(dhcp_host):
    """Return a host string for an address"""
    return "%s,%s.%s,%s" % (dhcp_host.mac_address,
                            dhcp_host.hostname,
                            FLAGS.linux_dhcp_domain,
                            dhcp_host.ip_address)

def _host_lease(dhcp_host):
    """Return a host string for an address in leasefile format"""
    seconds_since_epoch = calendar.timegm(dhcp_host.timestamp.utctimetuple())

    return "%d %s %s %s *" % (seconds_since_epoch + FLAGS.dhcp_lease_time,
                              dhcp_host.mac_address,
                              dhcp_host.address,
                              dhcp_host.hostname or '*')

def get_dhcp_hosts(dhcp_hosts):
    """Get a string containing a network's hosts config in dnsmasq format"""
    hosts = []
    for dhcp_host in dhcp_hosts:
        hosts.append(_host_dhcp(dhcp_host))
    return '\n'.join(hosts)

def get_dhcp_leases(dhcp_hosts):
    """Return a network's hosts config in dnsmasq leasefile format"""
    hosts = []
    for dhcp_host in dhcp_hosts:
        hosts.append(_host_lease(dhcp_host))
    return '\n'.join(hosts)

def update_dhcp(bridge, gateway, dhcp_start, dhcp_hosts, 
                dhcp_script=FLAGS.linux_dhcpbridge):
    """(Re)starts a dnsmasq server for a given network

    if a dnsmasq instance is already running then send a HUP
    signal causing it to reload, otherwise spawn a new instance
    """
    conffile = _dhcp_file(bridge, 'conf')
    with open(conffile, 'w') as f:
        f.write(get_dhcp_hosts(dhcp_hosts))

    # Make sure dnsmasq can actually read it (it setuid()s to "nobody")
    os.chmod(conffile, 0644)

    pid = _dnsmasq_pid_for(bridge)

    # if dnsmasq is already running, then tell it to reload
    if pid:
        out, _err = _execute('cat', "/proc/%d/cmdline" % pid,
                             check_exit_code=False)
        if conffile in out:
            try:
                _execute('sudo', 'kill', '-HUP', pid)
                return
            except Exception as exc:  # pylint: disable-msg=W0703
                LOG.debug(_("Hupping dnsmasq threw %s"), exc)
        else:
            LOG.debug(_("Pid %d is stale, relaunching dnsmasq"), pid)

    # FLAGFILE and DNSMASQ_INTERFACE in env
    env = {'FLAGFILE': FLAGS.linux_dhcpbridge_flagfile,
           'DNSMASQ_INTERFACE': bridge}
    command = _dnsmasq_cmd(bridge, gateway, dhcp_start, dhcp_script)
    _execute(*command, addl_env=env)

