# vim: tabstop=4 shiftwidth=4 softtabstop=4

# Copyright (c) 2011 NTT.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

""" Module that defines methods and classes relevant to the pluggable network
    service management."""

from nova import db
from nova import flags    
from nova import utils

from zope import interface

# Flag definitions
FLAGS = flags.FLAGS
flags.DEFINE_string('network_service_conf', 
                    '/etc/nova/nova-network-service.conf',
                    'Path to the network service configuration file.')
flags.DEFINE_string('default_network_service', 'nova.network.flat',
                    'Default network service to use.')
flags.DEFINE_bool('use_ipv6', False, 'use the ipv6')

class INetworkServiceFactory(interface.Interface):
    """A Network service factory object"""

    def get_net_agent():
        """Gets the Net agent service object.
        
        Returns:
           Net agent object.
        """
        
    def get_api_service():
        """Gets the API service object.
        
        Returns:
           Network API service object.
        """
        
    def get_os_api_service():
        """Gets the OpenStack API service object.
        
        Returns:
           Network OpenStack API service object.
        """

class INetworkApiService(interface.Interface):
    """A Network API service object."""
    
    def create_vnic():
        """Creates a new VNIC.
        
        Returns:
           The ID of the new VNIC.
        """
        
    def get_project_vpn_address_and_port(context, project_id):
        """Gets the VPN IP address and port for a project
        
        Args:
           context: Nova context object needed to access DB.
           project_id: Project to get the VPN IP and port for.
           
        Returns:
           A tuple of VPN IP address and port number.
        """

class INetworkOpenStackApiService(interface.Interface):
    """An OpenStack Network API service object."""
    
    def set_routes(route_map):
        """Set necessary routes for the plugin.
        
        Args:
           route_map: NetworkServiceRouteMap object to add the routes to.
        """ 

class INetworkAgentService(interface.Interface):
    """An OpenStack network agent service object."""
    
    def bind_vnics_to_ports(context, vnic_ids, is_vpn):
        """Gets a fixed ips from the pool.
        
        Args:
           context: Nova context needed for DB access.
           vnic_ids: list of VNIC IDs
           is_vpn: Boolean to check if the call is for VPN.          
        """
        
    def setup_compute_network(context, vnic_ids, is_vpn):
        """Set up the compute node for networking.
        
        Args:
           context: Nova context needed for DB access.
           vnic_ids: list of VNIC IDs
           is_vpn: Boolean to check if the call is for VPN.
        """
        
    def teardown_compute_network(context, vnic_ids):
        """Clean up the compute node.
        
        Args:
           context: Nova context needed for DB access.
           vnic_ids: list of VNIC IDs
        """
        
    def requires_file_injection():
        """Indicates whether the plugin requires file injection.
        
        Returns:
           True if the plugin requires file injection.
        """
        
    def get_network_info(context, vnic_id):
        """Gets network data for a given VNIC.
        
        Args:
           context: Nova context needed for DB access.
           vnic_id: VNIC ID to get the network information for.
        
        Returns:
           A dictionary containing the following keys:
           cidr, cidrv6, netmask, netmask_v6 gateway, gateway_v6, 
           dhcp_server, broadcast, dns, ra_server, mac_address,
           ip_address, bridge, and address_v6
        """

class NetworkServiceRouteMap(object):
    """Wrapper class for route mapper."""

    def __init__(self, mapper, service):
        """Initialize a NetworkServiceRouteMap object.
        
        Args:
           mapper: python-routes mapper object to wrap.
           service: module of the network plugin service.
        """
        self._mapper = mapper
        self._service = service

    def resource(self, member_name, collection_name, **kwargs):
        """Wrapper method for 'resource' method in python-routes mapper object.
        
        This method does the same thing 'resource' method does for python-routes
        module, but it also adds a route prefix from the network service name. 
        
        Args:
           member_name: REST member routes
           collection_name: REST collection routes
        """
        _, _, package_name = str(self._service.__package__).rpartition('.')
        kwargs.pop('path_prefix', None)
        self._mapper.resource(member_name, collection_name, 
                              path_prefix='/%s' % package_name, **kwargs)

class NetworkServiceManager(object):
    """Implements Borg design pattern to have a shared state of network 
    services.
    """

    __shared_state = {}
    _services = {}
    
    def __init__(self):
        """Initialize services member dictionary variable.
        """
        self.__dict__ = self.__shared_state

    def _import_services(self):
        """Loads the content of the configuration file and imports the 
        modules.
        """
        self._services = {}
        f = open(FLAGS.network_service_conf, 'r')
        for line in f:
            line = line.rstrip()
            if len(line) == 0 or line[0] == '#':
                continue
            self._services[line] = utils.import_object(line)
        f.close()
        
    @property
    def _services_dict(self):
        """Gets the services dictionary.  Loads the services if empty.
        
        Returns:
           A dictionary of service name to service modules.
        """
        if not self._services:
            self._import_services()
        return self._services

    @classmethod
    def get_service_name(cls, context, project_id):
        """Gets the network service given a project ID.
        
        Args:
           context: Nova context object needed for DB access.
           project_id: Project ID to get the service of.
           
        Returns:
           Network service name for the project.
        """
        module = db.api.project_get_network_service(context, project_id)
        if not module:
            module = FLAGS.default_network_service
        return module

    def get_service(self, context, project_id):
        """Gets the network service given a project ID.
        
        Args:
           context: Nova context object needed for DB access.
           project_id: Project ID to get the service of.
           
        Returns:
           Network service module for the project.
        """
        service = self.get_service_name(context, project_id)
        return self._services_dict[service]
        
    def get_all_services(self):
        """Gets all the network services available.
        
        Returns:
           Network service modules available.
        """
        return self._services_dict.values()

def get_service_factory(context, project_id):
    """Gets the factory object for network service.
    
    Args:
       context: Nova context needed for accessing DB.
       project_id: Project ID to get the appropriate network service.
       
    Returns:
       A tuple of the service module and its NetworkServiceFactory object.
    """
    manager = NetworkServiceManager()
    service = manager.get_service(context, project_id)
    return (service, service.NetworkServiceFactory())
    
def get_all_service_factories():
    """Gets all the factory objects for network services.
    
    Returns:
        A list of tuples of service module and its factory.
    """
    manager = NetworkServiceManager()
    services = manager.get_all_services()
    factories = []
    for service in services:
        factories.append((service,service.NetworkServiceFactory()))
    return factories

