# vim: tabstop=4 shiftwidth=4 softtabstop=4

# Copyright (c) 2011 NTT.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import socket
from zope import interface

from nova import flags
from nova import rpc
from nova import utils
from nova.network import service
from nova.network.vlan import common
from nova.network.vlan.db import api as db_api

FLAGS = flags.FLAGS
flags.DEFINE_string('ns_vlan_network_host', socket.gethostname(),
                    'Network host to use for ip allocation in flat modes')
flags.DEFINE_string('ns_vlan_network_topic', 'net-vlan', 
                    'the topic network nodes listen on')
flags.DEFINE_string('ns_vlan_network_bridge', 'br100',
                    'Bridge for simple network instances')
flags.DEFINE_bool('ns_vlan_network_fake_call', False,
                  'If True, skip using the queue and make local calls')
flags.DEFINE_string('ns_vlan_network_agent_driver', 
                    'nova.network.linux_net',
                    'Driver to use for network creation')
flags.DEFINE_string('ns_vlan_agent_interface', None,
                    'FlatDhcp will bridge into this interface if set')

def _queue_get_for(topic, physical_node_id):
    return "%s.%s" % (topic, physical_node_id)

def _get_network_topic(context, **kwargs):
    """Retrieves the network host for a project on this host"""
    if FLAGS.stub_network:
        host = FLAGS.ns_vlan_network_host
    else:
        host = _get_network_host(context)
    return _queue_get_for(FLAGS.ns_vlan_network_topic,
                          host)

def _set_network_host(context, network_id):
    """Safely sets the host of the network."""
    db_api.network_update(context,
                          network_id,
                          {"host":FLAGS.host,
                           "dns": FLAGS.ns_vlan_dns})
    return FLAGS.host 

def _get_network_host(context):
    """Get the network host for the current context."""
    network_ref = db_api.network_get_by_project(context,
                                                context.project_id)
    host = network_ref['host']
    if not host:
        topic = _queue_get_for(FLAGS.ns_vlan_network_topic,
                               FLAGS.ns_vlan_network_host)
        if FLAGS.ns_vlan_network_fake_call:
            return _set_network_host(context, network_ref['id'])
        host = rpc.call(context,
                        FLAGS.ns_vlan_network_topic,
                        {"method": "set_network_host",
                         "args": {"network_id": network_ref['id']}})
    return host

class NetworkAgentService(object):
    """An OpenStack network agent service object."""

    interface.implements(service.INetworkOpenStackApiService)

    def bind_vnics_to_ports(self, context, vnic_ids, is_vpn): 
        """Gets a fixed ips from the pool.
        
        Args:
           context: Nova context needed for DB access.
           vnic_ids: list of VNIC IDs
           is_vpn: Boolean to check if the call is for VPN.          
        """
        rpc.call(context,
                 _get_network_topic(context),
                 {"method": "allocate_fixed_ips",
                  "args": {"vnic_ids": vnic_ids,
                           "vpn": is_vpn}})
    
    def setup_compute_network(self, context, vnic_ids, is_vpn): 
        """Set up the compute node for networking.
        
        Args:
           context: Nova context needed for DB access.
           vnic_ids: list of VNIC IDs
           is_vpn: Boolean to check if the call is for VPN.
        """
        driver = utils.import_object(FLAGS.ns_vlan_network_agent_driver)
        for vnic_id in vnic_ids:
            network_ref = db_api.network_get_by_ethernet_card(context, vnic_id)
            driver.ensure_vlan_bridge(network_ref['vlan'], 
                                      network_ref['bridge'])
    
    def teardown_compute_network(self, context, vnic_ids):
        """Clean up the compute node.
        
        Args:
           context: Nova context needed for DB access.
           vnic_ids: list of VNIC IDs
        """
        for vnic_id in vnic_ids:
            fixed_ip = vnic_id.get('fixed_ip')
            if not FLAGS.stub_network and fixed_ip:
                floating_ips = fixed_ip.get('floating_ips') or []
    
                for floating_ip in floating_ips:
                    address = floating_ip['address']
                
                    network_topic = _queue_get_for(context,
                                         FLAGS.ns_vlan_network_topic,
                                         floating_ip['host'])
                
                    rpc.cast(context,
                             network_topic,
                             {"method": "disassociate_floating_ip",
                              "args": {"floating_address": address}})
                
                address = fixed_ip['address']
                if address:
                    # NOTE(vish): Currently, nothing needs to be done on the
                    #             network node until release. If this changes,
                
                    db_api.fixed_ip_update(context.elevated(), address,
                                           {'allocated':False}) 
    
    def requires_file_injection(self):
        """This service does not requires file injection.
        """
        return False
    
    def get_network_info(self, context, vnic_id):
        """Gets network data for a given VNIC.
        
        Args:
           context: Nova context needed for DB access.
           vnic_id: VNIC ID to get the network information for.
        
        Returns:
           A dictionary containing the following keys:
           cidr, cidrv6, netmask, netmask_v6 gateway, gateway_v6, 
           dhcp_server, broadcast, dns, ra_server, mac_address,
           ip_address, bridge, and address_v6
        """
        ethernet_card = db_api.ethernet_card_get(context, vnic_id)
        if not ethernet_card:
    	    raise ValueError("Invalid vnic ID!")
    
        # Get the IP address associated with this VNIC
        ip_address = db_api.ip_address_get_by_ethernet_card(context, 
    							ethernet_card.id)
        if not ip_address:
    	    raise ValueError("VNIC is not mapped to IP!")
    
        network = ip_address.network
        if network['cidr_v6']:
    	   ipv6_addr = utils.to_global_ipv6(network['cidr_v6'],
    					 ethernet_card['mac_address'])
        else:
    	   ipv6_addr = None
    
        net, mask = common.get_net_and_mask(network['cidr'])
        return dict(mac_address=ethernet_card['mac_address'],
    		ip_address=ip_address['address'], bridge=network['bridge'],
    		cidr=network['cidr'], netmask=network['netmask'],
    		gateway=network['gateway'], broadcast=network['broadcast'],
    		dns=network['dns'], dhcp_server=network['gateway'], 
    		cidr_v6=network['cidr_v6'], address_v6=ipv6_addr, 
    		ra_server=network['ra_server'], 
            netmask_v6=network['netmask_v6'], 
            gateway_v6=network['gateway_v6'],
    		net=net, mask=mask) 
