# vim: tabstop=4 shiftwidth=4 softtabstop=4

# Copyright (c) 2011 NTT.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Wrapper APIs for DB models should be defined here.
"""
import IPy
import math

from nova import exception
from nova import flags
from nova import utils

from nova.network.vlan.db import api

FLAGS = flags.FLAGS
flags.DEFINE_integer('ns_vlan_start', 100, 'First VLAN for private networks')
flags.DEFINE_integer('ns_vlan_vpn_start', 1000, 
                     'First Vpn port for private networks')

BOTTOM_RESERVED_IPS = 2
TOP_RESERVED_IPS = 1

def _create_fixed_ips(context, network_id):
    """Create all fixed ips for network."""
    network_ref = api.network_get(context, network_id)
    project_net = IPy.IP(network_ref['cidr'])
    num_ips = len(project_net)
    for index in range(num_ips):
        address = str(project_net[index])
        if index < BOTTOM_RESERVED_IPS or num_ips - index < TOP_RESERVED_IPS:
            reserved = True
        else:
            reserved = False
        api.fixed_ip_create(context, {'network_id': network_id,
                                      'address': address,
                                      'reserved': reserved})

def network_create(context, cidr, num_networks, network_size,
                   cidr_v6, vlan_start, vpn_start, **kwargs):
    """Create networks based on parameters."""
    # Check that num_networks + vlan_start is not > 4094, fixes lp708025
    if num_networks + vlan_start > 4094:
        raise ValueError(_('The sum between the number of networks and'
                           ' the vlan start cannot be greater'
                           ' than 4094'))

    fixed_net = IPy.IP(cidr)
    if fixed_net.len() < num_networks * network_size:
        raise ValueError(_('The network range is not big enough to fit '
              '%(num_networks)s. Network size is %(network_size)s' %
              locals()))

    fixed_net_v6 = IPy.IP(cidr_v6)
    network_size_v6 = 1 << 64
    significant_bits_v6 = 64
    networks = []
    for index in range(num_networks):
        vlan = vlan_start + index
        start = index * network_size
        start_v6 = index * network_size_v6
        significant_bits = 32 - int(math.log(network_size, 2))
        cidr = "%s/%s" % (fixed_net[start], significant_bits)
        project_net = IPy.IP(cidr)
        net = {}
        net['cidr'] = cidr
        net['netmask'] = str(project_net.netmask())
        net['gateway'] = str(project_net[1])
        net['broadcast'] = str(project_net.broadcast())
        net['vpn_private_address'] = str(project_net[2])
        net['dhcp_start'] = str(project_net[3])
        net['vlan'] = vlan
        net['bridge'] = 'br%s' % vlan
        if(FLAGS.use_ipv6):
            cidr_v6 = "%s/%s" % (fixed_net_v6[start_v6],
                                 significant_bits_v6)
            net['cidr_v6'] = cidr_v6

        # NOTE(vish): This makes ports unique accross the cloud, a more
        #             robust solution would be to make them unique per ip
        net['vpn_public_port'] = vpn_start + index
        network_ref = None
        try:
            network_ref = api.network_get_by_cidr(context, cidr)
        except exception.NotFound:
            pass

        if network_ref is not None:
            raise ValueError(_('Network with cidr %s already exists' %
                               cidr))

        network_ref = api.network_create_safe(context, net)
        if network_ref:
            networks.append(network_ref)
            _create_fixed_ips(context, network_ref['id'])

    return networks

def allocate_ip(context, ethernet_card_id, *args, **kwargs):
    """Gets an IP from the pool."""
    ctxt = context.elevated()
    network_ref = api.network_get_by_project(ctxt, context.project_id)

    if kwargs.get('vpn', None):
        ip_ref = api.fixed_ip_get_by_address(ctxt, 
                         network_ref['vpn_private_address'])
        api.fixed_ip_associate(ctxt, ip_ref['id'], ethernet_card_id)
    else:
        ip_ref = api.fixed_ip_associate_pool(ctxt,
                                             network_ref['id'],
                                             ethernet_card_id)

    api.fixed_ip_update(context, ip_ref['id'], {'allocated': True})
    return ip_ref 

def bind_ip_to_ethernet_card(context, ethernet_card_id):
    """Assigns IP to ethernet card."""
    
    # Get the ethernet card
    ethernet_card = api.ethernet_card_get(context, ethernet_card_id)
    # TODO: Throw a better exception.
    if not ethernet_card:
        raise ValueError("There is not VNIC with ID %s" % ethernet_card_id)

    ip_address = api.fixed_ip_get_by_ethernet_card(context, 
                                                   ethernet_card_id)
    # Check to see if this ethernet card does not have IP allocated.
    if not ip_address:
        ip_address = allocate_ip(context, ethernet_card_id)

    return ip_address

def ethernet_card_create_with_random_mac(context):
    """Creates a new ethernet card"""
    mac_address = utils.generate_mac()
    return api.ethernet_card_create(context, dict(mac_address=mac_address))
