# vim: tabstop=4 shiftwidth=4 softtabstop=4

# Copyright (c) 2011 NTT.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""
SQLAlchemy models for VLAN network service data.
"""

from sqlalchemy.orm import relationship, backref
from sqlalchemy import Column, Boolean, Integer, String
from sqlalchemy import ForeignKey
from sqlalchemy.ext.declarative import declarative_base

from nova import exception
from nova.db.sqlalchemy.models import NovaBase
from nova.network.flat_dhcp.db.sqlalchemy import models as flat_dhcp_models



BASE = declarative_base()


# ----------------------
# Data Transfer Objects.
# ----------------------

class EthernetCard(BASE, NovaBase):
    """Represents an ethernet card."""
    __tablename__ = 'ethernet_cards'
    id = Column(Integer, primary_key=True)
    mac_address = Column(String(255), nullable=False, unique=True)


class Network(BASE, NovaBase):
    """Represents a simple IP network."""
    __tablename__ = 'networks'
    id = Column(Integer, primary_key=True)
    cidr = Column(String(255))
    netmask = Column(String(255))
    bridge = Column(String(255))
    gateway = Column(String(255))
    broadcast = Column(String(255))
    dns = Column(String(255))
    vlan = Column(Integer)
    vpn_public_address = Column(String(255))
    vpn_public_port = Column(Integer)
    vpn_private_address = Column(String(255))
    dhcp_start = Column(String(255))
    cidr_v6 = Column(String(255), unique=True)
    ra_server = Column(String(255))
    label = Column(String(255))
    project_id = Column(String(255))
    host = Column(String(255))  # , ForeignKey('hosts.id'))


class FixedIp(BASE, NovaBase):
    """Represents a fixed ip for an instance."""
    __tablename__ = 'fixed_ips'
    id = Column(Integer, primary_key=True)
    address = Column(String(255))
    network_id = Column(Integer, ForeignKey('networks.id'), nullable=True)
    network = relationship(Network, backref=backref('fixed_ips'))
    ethernet_card_id = Column(Integer, ForeignKey('ethernet_cards.id'))
    ethernet_card = relationship(EthernetCard, backref=backref('fixed_ips'))
    allocated = Column(Boolean, default=False)
    leased = Column(Boolean, default=False)
    reserved = Column(Boolean, default=False)
    addressV6 = Column(String(255))
    netmaskV6 = Column(String(3))
    gatewayV6 = Column(String(255))


class FloatingIp(BASE, NovaBase):
    """Represents a floating ip that dynamically forwards to a fixed ip."""
    __tablename__ = 'floating_ips'
    id = Column(Integer, primary_key=True)
    address = Column(String(255))
    fixed_ip_id = Column(Integer, ForeignKey('fixed_ips.id'), nullable=True)
    fixed_ip = relationship(FixedIp,
                            backref=backref('floating_ips'),
                            foreign_keys=fixed_ip_id,
                            primaryjoin='and_('
                                'FloatingIp.fixed_ip_id == FixedIp.id,'
                                'FloatingIp.deleted == False)')
    project_id = Column(String(255))
    host = Column(String(255))  # , ForeignKey('hosts.id'))    


# ----------------------
# Data Access Objects.
# ----------------------

class VlanNetworkDataAccess(flat_dhcp_models.FlatDhcpNetworkDataAccess):
    """A Data Access Object to access all data for the VLAN network
    service.

    Also access fixed IP address associations.
    """

    ethernet_card_dto_class = EthernetCard

    network_dto_class = Network

    fixed_ip_dto_class = FixedIp

    floating_ip_dto_class = FloatingIp

    def network_get_by_project(self, project_id, associate=True):
        result = self._session.query(self.network_dto_class).\
            filter_by(project_id=project_id).\
            filter_by(deleted=False).\
            first()
        if not result:
            if not associate:
                return None
            try:
                return self.network_associate(project_id)
            except IntegrityError:
                # NOTE(vish): We hit this if there is a race and two
                #             processes are attempting to allocate the
                #             network at the same time
                result = self._session.query(self.network_dto_class).\
                    filter_by(project_id=project_id).\
                    filter_by(deleted=False).\
                    first()
        return result

    def network_get_by_cidr(self, cidr):
        if not self._is_admin:
            raise exception.NotAuthorized()
        result = self._session.query(self.network_dto_class).\
            filter_by(cidr=cidr).first()

        if not result:
            raise exception.NotFound(_('Network with cidr %s does not exist') %
                                     cidr)
        return result

    def network_associate(self, project_id):
        if not self._is_admin:
            raise exception.NotAuthorized()
        with self._session.begin():
            network_ref = self._session.query(self.network_dto_class).\
                filter_by(deleted=False).\
                filter_by(project_id=None).\
                with_lockmode('update').\
                first()
            # NOTE(vish): if with_lockmode isn't supported, as in sqlite,
            #             then this has concurrency issues
            if not network_ref:
                raise nova_api.NoMoreNetworks()
            network_ref['project_id'] = project_id
            self._session.add(network_ref)
        return network_ref

    def fixed_ip_get_network(self, address):
        if not self._is_admin:
            raise exception.NotAuthorized()
        fixed_ip_ref = self.fixed_ip_get_by_address(address)
        return fixed_ip_ref.network
